#
# ir.core - consumer infrared core infrastructure.
#
# Note that the root package (ir) pulls the interesting names directly
# into its namespace. You should not need to crawl in here unless you're
# messing with the internals.
#
# Copyright 2010-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import time
import string
import asyn
from functools import reduce


#
# System parameters.
#
PULSE_ABSOLUTE_TOLERANCE = 3	# pulse comparison tolerance (count)
PULSE_RELATIVE_TOLERANCE = 0.05	# pulse comparison tolerance (relative)
CARRIER_TOLERANCE = 0			# tolerance for carrier frequency comparison (Hz)

def tolerance(a, widen=0):
	return max(PULSE_ABSOLUTE_TOLERANCE, a * PULSE_RELATIVE_TOLERANCE, widen)

def close_enough(a, b, widen=0):
	if a is None or b is None: return True	# wildcard match
	return abs(a - b) <= tolerance(a, widen)


#
# Common error conditions.
# Note that in asyn-based components, these are passed around as tokens, rather than raised.
#
class Error(Exception):
	""" common exception base class for ir errors """
	pass


#
# A single (on, off) IR pulse.
# We turn the sender on for (on) ticks, then leave it off for (off) ticks, of the carrier clock.
# The carrier frequency is implied at this level, so we have no notion of wall time.
# Comparison is fuzzy up to TOLERANCE ticks.
# An off value of None is a wildcard.
#
class Pulse(object):
	""" A single (on,off) IR pulse pair

	A Pulse consists of the carrier wave being sent for "on" ticks of the carrier, followed
	by "off" ticks of no transmission. The Pulse object knows nothing about the value of
	the carrier frequency - in here, it's all in terms of "ticks", not wall clock time.

	off == None indicates indefinite duration, as produced by IR listeners for the last
	Pulse of a pulsetrain. When compared, this matches any off length.
	"""

	def __init__(self, on, off):
		on = int(on)
		off = int(off) if off else None
		(self.on, self.off) = (on, off)

	def ticks(self):
		return self.on + (self.off or 0)

	def __eq__(self, other, widen=0):
		""" Loose Pulse comparison with extra widening for those really sloppy situations. """
		return close_enough(self.on, other.on) and close_enough(self.off, other.off, widen)

	def __ne__(self, other): return not (self == other)

	def __hash__(self):
		return id(self)

	def edit(self, new_off):
		return Pulse(self.on, new_off)

	def form(self, floor=0):
		return "%d,%d" % (max(self.on, floor), max(self.off, floor))

	def __str__(self):
		if self.off is None:
			return '%d,*' % self.on
		else:
			return self.form()

	def __repr__(self):
		return "(%s)" % str(self)


def ticks(pulsetrain):
	return reduce(lambda x, y: x+y, map(Pulse.ticks, pulsetrain))


#
# IR classification types
# Types themselves are duck-typed. Anything can be an IR Type as long as it has:
#	type.name			Unique name of type (string)
#	type.exactMatch		True if Codes/Symbols of this type are exact values; False if fuzzy
#	type.claim(event)	Return a Code for an Event if it's yours, or None to decline
# In particular, most IR Types are modules.
#
TYPELIST = []
TYPEMAP = { }

def registerStandardType(type):
	TYPELIST.append(type)
	TYPEMAP[type.name] = type

def registerType(type):
	TYPELIST = [type] + TYPELIST
	TYPEMAP[type.name] = type


#
# A generic IR event
#
class Event(object):
	""" An actually observed IR Event.

		Event represents an actual IR observation. It contains a carrier frequency,
		a list of Pulses (the pulsetrain), and some ancilliary stuff.

		Events are built incrementally by some receiver object - usually an actual
		IR receiver, but possibly a synthetic generator of some sort.
	"""
	def __init__(self):
		""" Create a blank Event, ready to fill in. """
		self.carrier = None
		self.pulsetrain = []
		self.trailer = None
		self.inception = time.time()

	def set_carrier(self, carrier):
		""" Set the carrier frequency. """
		assert self.carrier is None
		self.carrier = carrier

	def add_pulse(self, pulse):
		""" Add a new pulse to this IREvent. """
		self.pulsetrain.append(pulse)

	def finish(self):
		""" Finish construction of this IREvent. """
		last = self.pulsetrain[-1]
		self.trailer = last.off
		self.pulsetrain[-1] = last.edit(None)		# clear final pulse timeout

	def set_contents(self, in_carrier, in_pulses):
		""" Set contents in one fell swoop. """
		self.carrier = in_carrier
		self.pulsetrain = in_pulses
		self.finish()

	def is_complete(self):
		""" True if this event has been completely recorded. False if in progress. """
		return self.trailer is not None

	def __getitem__(self, *args):
		return self.pulsetrain.__getitem__(*args)

	def __eq__(self, other):
		return self.carrier == other.carrier and self.pulsetrain == other.pulsetrain

	def __ne__(self, other): return not self == other

	def __str__(self):
		""" Canonical string form of an Event. Vendor agnostic. """
		if self.is_complete():
			return "<%d,%s:%d>" % (self.carrier, ','.join(map(str, self.pulsetrain)), self.trailer)
		else:
			return "<%d,%s...>" % (self.carrier, ','.join(map(str, self.pulsetrain)))

	def __repr__(self):
		return str(self)

	def ticks(self):
		return ticks(self.pulsetrain)

	def duration(self):
		""" length of transmission, in seconds. """
		return float(self.ticks()) / self.carrier

	def matches(self, pattern):
		""" Returns true if self matches pattern. Partial events match "so far".

			If we are being built, match the Pulses received so far - this may be a partial match.
			If we're done, require a full match.
			Returns non-false if matched; or false if not.
		"""
		# There's GOT to be a better way to do this...
		mylen = len(self.pulsetrain)
		blen = len(pattern.base)
		if mylen < blen:
			if self.is_complete(): return False
			if self.pulsetrain != pattern.base[0:mylen]: return False
			return True
		if mylen == blen:
			# Careful, now: this may be a full run, with pattern[-1] not yet wildcarded
			if self.pulsetrain[:-1] != pattern.base[:-1]: return False
			return True
		if not sig.repeat: return False
		if self.pulsetrain[0:blen] != pattern.base: return False	# mismatch base
		rlen = len(pattern.repeat)
		(rep, rem) = divmod(mylen - blen, rlen)
		if rem and self.is_complete(): return False				# not a full repetition
		for p in range(blen, mylen-rlen, rlen):
			if self.pulsetrain[p:p+rlen] != pattern.repeat: return False # repeat mismatch
		if rem:
			if self.pulsetrain[-rem:] != pattern.repeat[0:rem]: return False # final partial mismatch
			return True
		return True

	def classify(self):
		""" Find an IR type that recognizes this Event.

			Run through the registered IR types in order, and take
			the first one who is willing to produce a Code for it.
			If none are willing, return None. (This will only happen
			if you don't have the 'generic' type in your list.)
		"""
		for type in TYPELIST:
			code = type.claim(self)
			if code:
				return code
		return None


#
# Superclass of all IR receiver objects.
# Doesn't do anything interesting yet.
#
class Receiver(object):
	""" A receiver (factory) of actually observed IR signals. """


#
# IR types are singleton modules, so we can't pickle them.
# Instead of descending into __reduce__ madness, we just store their names.
#
class _Typed(object):

	def __init__(self, type):
		self._type = type.name

	def typename(self):
		return self._type

	def type(self):
		return TYPEMAP[self._type]


#
# A *pattern* to match against, expressed in a particular type scheme.
#
class Symbol(_Typed):
	""" A meaningful IR pattern based on a particular type. """

	def __init__(self, type):
		_Typed.__init__(self, type)

	def symbol(self):
		""" For consistency, allow a Symbol to return itself as "its" symbol. """
		return self

	_CANONICALFORM = re.compile(r'<([a-zA-Z]+):(.*)>', re.I|re.S)
	@staticmethod
	def from_text(text):
		""" Convert canonical text form to a Symbol.

			This is primarily meant to reverse str(Symbol), so that
			Symbols can be stored in arbitrary string containers outside
			our code. As a special indulgence, we accept raw pronto hex-code
			as a valid representation of a generic Symbol.
			Note that str->from_text is usually the identity, but the reverse
			is not true since from_text will mercilessly canonicalize the Symbol.
		"""
		if not text:		# empty
			return None
		match = Symbol._CANONICALFORM.match(text)
		if match:
			typename, typespec = match.group(1, 2)
			if typename in TYPEMAP:
				return TYPEMAP[typename].Symbol.from_text(typespec)
			else:
				return None
		if text[0] in string.digits:	# assume hex-form string
			from ir.hex import decodeSymbol as _hexDecode # (break import loop at rt)
			return _hexDecode(text)
		# no dice
		return None


#
# An actual transmission, expressed in a particular type scheme.
#
class Code(_Typed):
	""" A concrete IREvent, interpreted as a particular IRType and matched to an IRSymbol. """

	def __init__(self, type):
		_Typed.__init__(self, type)

	def code(self):
		""" For consistency, allow a Code to return itself as "its" code. """
		return self
