#
# ir_hex - PRONTO-style "hex" encoding support
#
# "Hex" format (type zero, the only one we support) records base
# and repeat runs ("burst pair sequences"). It's pretty close to
# our notion of Symbol (not Code or Event). The only difference
# is that pure repeat patterns have empty base in hex, while we
# use base == repeat for this case.
#
# This is heavily rigged for sending hex input through the classifier.
# Don't fight it.
#
# Copyright 2010-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ir


HEADER_COUNT = 4
TYPE_RAW = 0
CARRIER_INV = 4145146
def pronto_carrier(x): return 1000000 / (x * 0.241246)		# frequency from divider (pronto magic)


class Error(ir.Error):
	""" Errors arising from importing Pronto "hex" codes. """


#
# Symbol -> hex form
#
def encode(symbol):
	""" Produce the Pronto "hex" form of a Symbol. """
	(base, repeat, count) = symbol.code(3).form()		# 3 is arbitrary; anything >1 will do
	assert count == 3									# but check that we got what we asked for
	if base == repeat:	# pure repeat pattern
		base = []		# hex format uses empty base for this
	out = "0000 %04x %04x %04x " % (
		CARRIER_INV / symbol.carrier,
		len(base),
		len(repeat) if repeat else 0
	)

	def hexes(train):
		""" Convert a pulse train to hex form, dropping any trailing item. """
		return " ".join(["%04x %04x" % (p.on, p.off)
			for p in [pulse for pulse in train if pulse.off]
		])

	out += hexes(base)
	if repeat:
		out += " " + hexes(repeat)
	return out


#
# Hex form -> Event
#
def decode(hex_string, repeats=3):
	""" Convert a pronto-style hex stream into an ir.Event.

		ir.Event is "rawer" than pronto in that the pronto format has identified
		any repeating part of the data. By default, we arbitrarily produce an event with
		three repetitions, which is suitable to be fed to ir.classify for re-analysis.
		If you'd rather produce an Event for direct emission, give your own repeat
		count as an argument.
	"""
	return HexEvent(hex_string, repeats)


#
# Straight conversion from hex to Symbol
#
def decodeSymbol(hex_string):
	""" Convert a pronto-style hex stream into an ir.Symbol.

		Here is the fun part: Pronto format is Symbol-like.
		Still, we convert it to Event level (with an arbitrary repeat
		of three) so we can feed it into the classifier. Worst case,
		the generic type will pick it up and filter the repeats back out
		for us. Better case, one of the specific recognizers gets it.
	"""
	return decode(hex_string).classify().symbol()


class HexEvent(ir.Event):
	""" An ir.Event generated from a pronto "hex" string. """

	def __init__(self, hex_string, repeats):
		ir.Event.__init__(self)
		try:
			hexes = map(lambda h: int(h, 16), hex_string.split())
		except ValueError:
			raise Error("invalid format")
		if len(hexes) < HEADER_COUNT + 2: raise Error("too short")

		# 4*short header: [format] [carrier] [] []
		if hexes[0] != TYPE_RAW: raise Error("not in raw format")
		carrier = round(CARRIER_INV/hexes[1], -3)	# rounded to nearest kHz
		r1c = hexes[2]
		r2c = hexes[3]

		# extract pulsetrains ("burst pair sequences" in Pronto-speak)
		def make_run(chain):
			return [ir.Pulse(*p) for p in zip(chain[0::2], chain[1::2])]
		run1 = make_run(hexes[HEADER_COUNT:HEADER_COUNT+r1c*2])
		run2 = make_run(hexes[HEADER_COUNT+r1c*2:])
		self.set_contents(carrier, run1 + run2 * repeats)
