#
# Global Cache interface plugin for Indigo 5
#
# Copyright 2011-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import indigo
import asyn
import cyin
import cyin.devstate
import ir
import ir.gcache.irl
import serial
from cyin import debug, error
from cyin.asynplugin import action
from cyin.check import *

import cyin.debugging


#
# A ConfigUI class with a "Learn" button that fills in the "ircode" field
# (which needs to exist).
#
LEARNING_TIMEOUT = 9.5	# 10s timeout imposed by Indigo - we must recover before then

class LearningUI(cyin.ConfigUI):
	@cyin.button
	def learn_symbol(self):
		learner = cyin.plugin.learner
		if learner:
			if learner.ready():
				debug("learning ir from", learner.name)
				try:
					code = cyin.plugin.inject_callout(learner.learn_symbol,
						timeout=LEARNING_TIMEOUT,
						timeout_notify=lambda: learner.learn_symbol(reply=None))
					if code:
						self.ircode = cyin.plugin.inject_wait(cyin.plugin.to_string, code.symbol())
				except Exception as e:
					error(learner.name, "learning aborted:", e)
			else:
				error(learner.name, "not ready - cannot learn from it")
		else:
			error("no learning device configured")


#
# A ConfigUI check function for IR symbol strings
#
def check_symbol(s):
	""" Checkrule for strings denoting IR symbols. """

	def validate(s):
		try:
			symbol = cyin.plugin.from_string(s)
			if symbol is None:
				return ('Invalid IR symbol',)
			return cyin.plugin.to_string(symbol)
		except Exception as e:
			return ('Invalid IR symbol (%s)' % e.message,)

	return cyin.plugin.inject_wait(validate, s)


#
# Common features of GCNet device users
#
class _GCDev(cyin.devstate.SubDevice):
	""" Common features of devices possibly attached to a GCNet host. """

	gcdev = None

	def set_subaddress(self, subaddr):
		self.devaddr = subaddr

	def stop(self):
		if self.gcdev:
			self.gcdev.remove_callout(self._calldown, required=False)
		super(_GCDev, self).stop()

	def get_gcdev(self):
		gcnet = self.hostdev.gcnet
		if gcnet and gcnet.devmap:
			if self.devaddr in gcnet.devmap:
				self.set_display_address("%s,%s" % (self.hostdev.name, self.devaddr))
				return gcnet.devmap[self.devaddr]
			self.fail_hard("no device %s in %s" % (self.devaddr, self.hostdev.name))
		return None		# not yet (call back when hostdev is ready)


#
# An IR emitter attached to a GC device port
#
class IREmitter(_GCDev):
	""" An IR emitter device attached to a GC host. Can send IR signals. """

	queueDepth = cyin.DeviceState(name="queue", type=int) # transmit queue length (0 -> idle)

	def setup(self, ctx=None):
		self.gcdev = self.get_gcdev()
		if self.gcdev:
			self.gcdev.add_callout(self._calldown)
			self.proceed("ready", recovered=True)

	@action
	def sendIR(self, action):
		symbol = cyin.plugin.from_string(action.ircode)
		if symbol is None:
			return error("unable to parse IR symbol - check the action")
		debug("send", symbol, "to", self, "repeating", action.count)
		self.gcdev.emit(symbol, action.count)
		self.updateState()

	@action
	def clearIR(self, action):
		debug("clearIR not implemented")

	def _calldown(self, ctx, data=None):
		if ctx.state == 'transmitted':
			IRComplete.trigger(self, data)
			self.updateState()

	def updateState(self):
		self.queueDepth = len(self.gcdev.queue())


#
# An IR receiver attached to either a local or a GC-hosted serial port
#
class IRReceiver(_GCDev):
	""" An IR receiver attached to either a local or a GC-hosted serial port. """
	lastir = cyin.PluginProperty(name="lastir_var", required=False, reconfigure=False)
	lastir_var = cyin.NamedVariable(name="lastir")

	irl = None
	active_learner = None		# IR learner bypass

	def setup(self, ctx=None):
		if self.hostdev:	# network
			gcnet = self.get_gcdev()
			if gcnet:
				gcnet.connect(callout=self._connected)
		else:
			self.set_display_address(self.address)
			try:
				irl = serial.Serial(self.address, baudrate=ir.gcache.irl.BAUDRATE)
				self._connected(asyn.Context('serial'), irl)
			except serial.SerialException as e:
				return self.fail_soft("unable to open %s: %s" % (self.address, e))

	def _connected(self, ctx, irl=None):
		if ctx.error:
			return self.fail_soft(ctx.error)
		if ctx.state == 'connected':	# irl is a socket
			irl = irl.makefile("r+", 1)
		self.gcdev = self.irl = ir.gcache.irl.IRL(cyin.plugin.file(irl), callout=self._calldown)
		self.proceed("ready", recovered=True)

	def stop(self):
		_GCDev.stop(self)
		self.close()

	def close(self):
		if self.irl:
			self.irl.close()
		self.irl = None

	def learn_symbol(self, reply):
		self.active_learner = reply	# who to send our reply to

	def _calldown(self, ctx, event=None):
		if ctx.error:
			return self.fail_soft(ctx.error)
		code = event.classify()
		debug(self.name, "received", code)
		if self.active_learner:
			reply = self.active_learner
			self.active_learner = None
			return reply.callout('learned', code)
		if self.lastir:
			self.lastir_var = cyin.plugin.to_string(code)
		IRReceived.trigger(self, code)


#
# A relay device
#
class Relay(_GCDev):
	""" A relay device attached to a GC host. Can switch dry contacts. """

	onOffState = cyin.DeviceState(type=bool)

	def setup(self, ctx=None):
		self.gcdev = self.get_gcdev()
		if self.gcdev:
			self.gcdev.add_callout(self._calldown)
			self.proceed("ready", recovered=True)
			self.gcdev.poll()

	@action
	def standard_switch(self, active):
		self.gcdev.engage(active)

	@action
	def standard_toggle(self):
		self.gcdev.engage(not self.onOffState)

	@action
	def standard_status(self):
		self.gcdev.poll()

	def _calldown(self, ctx, data=None):
		if ctx.error:
			return self.fail_soft(ctx.error)
		if ctx.state == 'state':	# unreliable
			debug(self.name, "state reported", data)
			self.onOffState = data


#
# A relay device
#
class Sensor(_GCDev):
	""" A sensor device attached to a GC host. Can poll or notify. """

	onOffState = cyin.DeviceState(type=bool)

	def setup(self, ctx=None):
		self.gcdev = self.get_gcdev()
		if self.gcdev:
			self.gcdev.add_callout(self._calldown)
			self.proceed("ready", recovered=True)
			self.gcdev.poll()

	@action
	def standard_status(self):
		self.gcdev.poll()

	def _calldown(self, ctx, data=None):
		if ctx.error:
			return self.fail_soft(ctx.error)
		if ctx.state == 'state':
			debug(self.name, "state reported", data)
			self.onOffState = data


#
# IR-related events
#
class _IREvent(cyin.Event):
	""" Common event features of IR-related events. """

	device = cyin.PluginProperty(type=cyin.device, required=False)
	ircode = cyin.PluginProperty(required=False, check=[check_symbol])

	@cyin.cached
	def symbol(self):
		if self.ircode:
			return cyin.plugin.from_string(self.ircode)

	def matches(self, emitter, ircode):
		if self.device:
			if self.device != emitter:
				return False
		if self.symbol:
			if not self.symbol.matches(ircode):
				return False
		return True


class IRComplete(_IREvent):
	""" Event: An IR event has been successfully sent. """

class IRReceived(_IREvent):
	""" Event: we received an IR event. """
	UI = LearningUI


#
# Actions
#
class SendIR(cyin.DeviceAction):
	""" Action: send an IR signal to an emitter device. """
	ircode = cyin.PluginProperty(check=[check_symbol], eval=True)
	count = cyin.PluginProperty(type=int, eval=True, check=[check_int(min=0)])

	class UI(LearningUI):
		def description(self):
			return "send %s to %s" % (self.ircode, self.device.name)


class ClearIR(cyin.Action):
	""" Action: Clear the queue of an emitter device. """
	pass
